#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Creational API for LDF (LIN Description File) components.

This module provides a high-level API for creating and manipulating LDF components
such as LIN networks, master/slave nodes, frames, signals, and schedule tables.
"""

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2010-2023 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__ = "Christoph Schueler"
__version__ = "0.1.0"

from typing import Optional, List, Union, Tuple

from pydbc.db import VNDB
from pydbc.db.model import (
    LinMasterNode,
    LinNetwork,
    LinNode,
    LinUnconditionalFrame,
    LinSignal,
    LinSlaveNode,
    LinEventTriggeredFrame,
    LinScheduleTable,
    LinScheduleTable_Command_Frame,
    LinSporadicFrame,
    LinSignalEncodingType,
    LinSignalEncodingEntry_Logical,
    LinSignalEncodingEntry_Physical,
    LinSignalRepresentation,
    LinUnconditionalFrameSignal,
    LinSignalSubscriber,
)


class LDFCreator:
    """High-level API for creating LDF components."""

    def __init__(self, db_path: str = ":memory:", debug: bool = False):
        """Initialize a new LDF creator.

        Args:
            db_path: Path to the database file or ":memory:" for in-memory database
            debug: Enable debug mode
        """
        self.db = VNDB.create(db_path, debug=debug)
        self.session = self.db.session
        self._networks = {}
        self._nodes = {}
        self._master_nodes = {}
        self._slave_nodes = {}
        self._frames = {}
        self._signals = {}
        self._schedule_tables = {}
        self._signal_encodings = {}

    def create_network(
        self,
        name: str,
        protocol_version: Optional[str] = None,
        language_version: Optional[str] = None,
        speed: Optional[float] = None,
        file_revision: Optional[str] = None,
        channel_name: Optional[str] = None,
        **kwargs
    ) -> LinNetwork:
        """Create a new LIN network.

        Args:
            name: Name of the network
            protocol_version: LIN protocol version
            language_version: LIN language version
            speed: Network speed in kbps
            file_revision: File revision
            channel_name: Channel name
            **kwargs: Additional network attributes

        Returns:
            The created LinNetwork object
        """
        network = LinNetwork(
            name=name,
            protocol_version=protocol_version,
            language_version=language_version,
            speed=speed,
            file_revision=file_revision,
            channel_name=channel_name,
            **kwargs
        )
        self.session.add(network)
        self._networks[name] = network
        return network

    def create_master_node(
        self,
        name: str,
        timebase: float,
        jitter: float,
        bit_length: Optional[int] = None,
        tolerant: Optional[bool] = None,
        **kwargs
    ) -> LinMasterNode:
        """Create a new LIN master node.

        Args:
            name: Name of the node
            timebase: Timebase in seconds
            jitter: Jitter in seconds
            bit_length: Bit length
            tolerant: Whether the node is tolerant
            **kwargs: Additional node attributes

        Returns:
            The created LinMasterNode object
        """
        master_node = LinMasterNode(
            name=name,
            timebase=timebase,
            jitter=jitter,
            bit_length=bit_length,
            tolerant=tolerant,
            **kwargs
        )
        self.session.add(master_node)
        self._nodes[name] = master_node
        self._master_nodes[name] = master_node
        return master_node

    def create_slave_node(
        self,
        name: str,
        protocol_version: Optional[str] = None,
        configured_NAD: Optional[int] = None,
        initial_NAD: Optional[int] = None,
        product_id: Tuple[int, int, int] = (),
        p2_min: Optional[float] = None,
        st_min: Optional[float] = None,
        n_as_timeout: Optional[float] = None,
        n_cr_timeout: Optional[float] = None,
        response_tolerance: Optional[float] = None,
        **kwargs
    ) -> LinSlaveNode:
        """Create a new LIN slave node.

        Args:
            name: Name of the node
            protocol_version: LIN protocol version
            configured_NAD: Configured NAD
            initial_NAD: Initial NAD
            product_id: Product ID tuple (supplier_id, function_id, variant)
            p2_min: P2 minimum time
            st_min: ST minimum time
            n_as_timeout: N_As timeout
            n_cr_timeout: N_Cr timeout
            response_tolerance: Response tolerance
            **kwargs: Additional node attributes

        Returns:
            The created LinSlaveNode object
        """
        slave_node = LinSlaveNode(
            name=name,
            protocol_version=protocol_version,
            configured_NAD=configured_NAD,
            initial_NAD=initial_NAD,
            product_id=product_id,
            p2_min=p2_min,
            st_min=st_min,
            n_as_timeout=n_as_timeout,
            n_cr_timeout=n_cr_timeout,
            response_tolerance=response_tolerance,
            **kwargs
        )
        self.session.add(slave_node)
        self._nodes[name] = slave_node
        self._slave_nodes[name] = slave_node
        return slave_node

    def create_signal(
        self,
        name: str,
        signal_size: int,
        init_value: Union[int, List[int]],
        publisher: Union[str, LinNode],
        **kwargs
    ) -> LinSignal:
        """Create a new LIN signal.

        Args:
            name: Name of the signal
            signal_size: Size of the signal in bits
            init_value: Initial value (integer or list of integers)
            publisher: Publisher node (name or LinNode object)
            **kwargs: Additional signal attributes

        Returns:
            The created LinSignal object
        """
        if isinstance(publisher, str):
            publisher = self._nodes[publisher]

        signal = LinSignal(
            name=name,
            signal_size=signal_size,
            init_value=init_value,
            publisher=publisher,
            **kwargs
        )
        self.session.add(signal)
        self._signals[name] = signal
        return signal

    def add_signal_subscriber(
        self, signal: Union[str, LinSignal], subscriber: Union[str, LinNode]
    ) -> LinSignalSubscriber:
        """Add a subscriber to a signal.

        Args:
            signal: Signal name or LinSignal object
            subscriber: Subscriber node name or LinNode object

        Returns:
            The created LinSignalSubscriber object
        """
        if isinstance(signal, str):
            signal = self._signals[signal]
        if isinstance(subscriber, str):
            subscriber_obj = self._nodes[subscriber]
        else:
            subscriber_obj = subscriber
        # subscriber_obj = LinSignalSubscriber(subscriber=subscriber)
        signal.subscribers.append(subscriber_obj)
        self.session.add(subscriber_obj)
        return subscriber_obj

    def create_unconditional_frame(
        self,
        name: str,
        frame_id: int,
        size: int,
        publisher: Union[str, LinNode],
        **kwargs
    ) -> LinUnconditionalFrame:
        """Create a new LIN unconditional frame.

        Args:
            name: Name of the frame
            frame_id: Frame ID
            size: Frame size in bytes
            publisher: Publisher node (name or LinNode object)
            **kwargs: Additional frame attributes

        Returns:
            The created LinUnconditionalFrame object
        """
        if isinstance(publisher, str):
            publisher = self._nodes[publisher]

        frame = LinUnconditionalFrame(
            name=name, frame_id=frame_id, size=size, publisher=publisher, **kwargs
        )
        self.session.add(frame)
        self._frames[name] = frame
        return frame

    def add_signal_to_frame(
        self,
        frame: Union[str, LinUnconditionalFrame],
        signal: Union[str, LinSignal],
        signal_offset: int = 0,
    ) -> LinUnconditionalFrameSignal:
        """Add a signal to a frame.

        Args:
            frame: Frame name or LinUnconditionalFrame object
            signal: Signal name or LinSignal object
            signal_offset: Signal offset in bits

        Returns:
            The created LinUnconditionalFrameSignal object
        """
        if isinstance(frame, str):
            frame = self._frames[frame]
        if isinstance(signal, str):
            signal = self._signals[signal]

        frame_signal = LinUnconditionalFrameSignal(
            signal=signal, signal_offset=signal_offset
        )
        frame.frame_signals.append(frame_signal)
        self.session.add(frame_signal)
        return frame_signal

    def create_event_triggered_frame(
        self,
        name: str,
        frame_id: int,
        master_node: Union[str, LinMasterNode],
        collision_resolving_schedule_table: Optional[str] = None,
        **kwargs
    ) -> LinEventTriggeredFrame:
        """Create a new LIN event-triggered frame.

        Args:
            name: Name of the frame
            frame_id: Frame ID
            master_node: Master node (name or LinMasterNode object)
            collision_resolving_schedule_table: Name of collision resolving schedule table
            **kwargs: Additional frame attributes

        Returns:
            The created LinEventTriggeredFrame object
        """
        if isinstance(master_node, str):
            master_node = self._master_nodes[master_node]

        frame = LinEventTriggeredFrame(
            name=name,
            frame_id=frame_id,
            master_node=master_node,
            collision_resolving_schedule_table=collision_resolving_schedule_table,
            **kwargs
        )
        self.session.add(frame)
        self._frames[name] = frame
        return frame

    def create_sporadic_frame(self, name: str, **kwargs) -> LinSporadicFrame:
        """Create a new LIN sporadic frame.

        Args:
            name: Name of the frame
            **kwargs: Additional frame attributes

        Returns:
            The created LinSporadicFrame object
        """
        frame = LinSporadicFrame(name=name, **kwargs)
        self.session.add(frame)
        self._frames[name] = frame
        return frame

    def create_schedule_table(self, name: str, **kwargs) -> LinScheduleTable:
        """Create a new LIN schedule table.

        Args:
            name: Name of the schedule table
            **kwargs: Additional schedule table attributes

        Returns:
            The created LinScheduleTable object
        """
        schedule_table = LinScheduleTable(name=name, **kwargs)
        self.session.add(schedule_table)
        self._schedule_tables[name] = schedule_table
        return schedule_table

    def add_frame_to_schedule_table(
        self,
        schedule_table: Union[str, LinScheduleTable],
        frame: Union[str, LinUnconditionalFrame],
        frame_time: float,
    ) -> LinScheduleTable_Command_Frame:
        """Add a frame to a schedule table.

        Args:
            schedule_table: Schedule table name or LinScheduleTable object
            frame: Frame name or LinUnconditionalFrame object
            frame_time: Frame time in seconds

        Returns:
            The created LinScheduleTable_Command_Frame object
        """
        if isinstance(schedule_table, str):
            schedule_table = self._schedule_tables[schedule_table]
        if isinstance(frame, str):
            frame = self._frames[frame]

        command = LinScheduleTable_Command_Frame(frame_time=frame_time, frame=frame)
        schedule_table.commands.append(command)
        self.session.add(command)
        return command

    def create_signal_encoding_type(self, name: str, **kwargs) -> LinSignalEncodingType:
        """Create a new LIN signal encoding type.

        Args:
            name: Name of the signal encoding type
            **kwargs: Additional signal encoding type attributes

        Returns:
            The created LinSignalEncodingType object
        """
        encoding_type = LinSignalEncodingType(name=name, **kwargs)
        self.session.add(encoding_type)
        self._signal_encodings[name] = encoding_type
        return encoding_type

    def add_logical_value_to_encoding(
        self,
        encoding_type: Union[str, LinSignalEncodingType],
        signal_value: int,
        text_info: str,
    ) -> LinSignalEncodingEntry_Logical:
        """Add a logical value to a signal encoding type.

        Args:
            encoding_type: Signal encoding type name or LinSignalEncodingType object
            signal_value: Signal value
            text_info: Text description

        Returns:
            The created LinSignalEncodingEntry_Logical object
        """
        if isinstance(encoding_type, str):
            encoding_type = self._signal_encodings[encoding_type]

        entry = LinSignalEncodingEntry_Logical(
            signal_value=signal_value, text_info=text_info
        )
        encoding_type.entries.append(entry)
        self.session.add(entry)
        return entry

    def add_physical_range_to_encoding(
        self,
        encoding_type: Union[str, LinSignalEncodingType],
        min_value: float,
        max_value: float,
        scale: float,
        offset: float,
        text_info: str,
    ) -> LinSignalEncodingEntry_Physical:
        """Add a physical range to a signal encoding type.

        Args:
            encoding_type: Signal encoding type name or LinSignalEncodingType object
            min_value: Minimum value
            max_value: Maximum value
            scale: Scale factor
            offset: Offset value
            text_info: Text description

        Returns:
            The created LinSignalEncodingEntry_Physical object
        """
        if isinstance(encoding_type, str):
            encoding_type = self._signal_encodings[encoding_type]

        entry = LinSignalEncodingEntry_Physical(
            min_value=min_value,
            max_value=max_value,
            scale=scale,
            offset=offset,
            text_info=text_info,
        )
        encoding_type.entries.append(entry)
        self.session.add(entry)
        return entry

    def add_signal_representation(
        self,
        signal: Union[str, LinSignal],
        encoding_type: Union[str, LinSignalEncodingType],
    ) -> LinSignalRepresentation:
        """Associate a signal with a signal encoding type.

        Args:
            signal: Signal name or LinSignal object
            encoding_type: Signal encoding type name or LinSignalEncodingType object

        Returns:
            The created LinSignalRepresentation object
        """
        if isinstance(signal, str):
            signal = self._signals[signal]
        if isinstance(encoding_type, str):
            encoding_type = self._signal_encodings[encoding_type]

        representation = LinSignalRepresentation(
            signal_encoding_type=encoding_type, signal=signal
        )
        self.session.add(representation)
        return representation

    def commit(self):
        """Commit changes to the database."""
        self.session.commit()

    def close(self):
        """Close the database connection."""
        self.session.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2010-2020 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__  = 'Christoph Schueler'
__version__ = '0.1.0'


import enum

CAN_EXTENDED_IDENTIFIER = 0x80000000

class VndbType(enum.IntEnum):
    SINGLE_NETWORK = 0
    MULTIPLE_NETWORKS = 1


class BusType(enum.IntEnum):
    CAN = 0
    LIN = 1


class FileType(enum.IntEnum):
    DBC = 0
    LDF = 1
    NCF = 2

class AttributeType(enum.IntEnum):
    NODE        = 0
    MESSAGE     = 1
    SIGNAL      = 2
    ENV_VAR     = 3
    NETWORK     = 4
    REL_NODE    = 5
    REL_SIGNAL  = 6
    REL_ENV_VAR = 7


class ValueType(enum.IntEnum):
    INT = 0
    HEX = 1
    FLOAT = 2
    STRING = 3
    ENUM = 4


class ByteOrderType(enum.IntEnum):
    MOTOROLA = 0
    INTEL = 1


class SignalType(enum.IntEnum):
    SINT = 0
    FLOAT32 = 1
    FLOAT64 = 2
    UINT = 8


class MultiplexingType(enum.IntEnum):
    NONE = 0
    MULTIPLEXOR = 1
    DEPENDENT = 2


class IdentifierType(enum.IntEnum):
    STANDARD = 0
    EXTENDED = 1


class EnvVarType(enum.IntEnum):
    INT = 0
    FLOAT = 1
    STRING = 2
    DATA = 3


class EnvVarAccessType(enum.IntEnum):
    UNRESTRICTED = 0
    READ = 1
    WRITE = 2
    READ_WRITE = 3


class ValueTableType(enum.IntEnum):
    SIGNAL  = 0
    ENV_VAR = 1


class CategoryType(enum.IntEnum):
    NODE = 0
    MESSAGE = 1
    ENV_VAR = 2


EXTENDED_ID_MASK = 0x80000000


class AddressBaseType:

    def __init__(self):
        pass


class CANAddress(AddressBaseType):
    """
    """

    def __init__(self, rawId):
        if (rawId & EXTENDED_ID_MASK) == EXTENDED_ID_MASK:
            self.mtype = IdentifierType.EXTENDED
            self.value = rawId & (~0x80000000)
        else:
            self.mtype = IdentifierType.STANDARD
            self.value = mid = rawId

    def __str__(self):
        """
        """
        return "{}({:08x} [{}])".format(self.__class__.__name__, self.value, self.mtype.name)

    def __int__(self):
        """
        """
        value = self.mid
        if self.mtype == IdentifierType.EXTENDED:
            value |= EXTENDED_ID_MASK
        return value


    __repr__ = __str__


class J1939Address(AddressBaseType):
    """

    """

    def __init__(self, priority, reserved, datapage, pdu_format, pdu_specific, source_address):
        self.priority = priority
        self.reserved = reserved
        self.datapage = datapage
        self.pdu_format = pdu_format
        self.pdu_specific = pdu_specific
        self.source_address = source_address

    @classmethod
    def from_int(klass, canID):
        # TODO: check for extID
        canID &= (~CAN_EXTENDED_IDENTIFIER)
        priority = (canID & 0x1c000000) >> 26
        reserved = (canID & 0x2000000) >> 25
        datapage = (canID & 0x1000000) >> 24
        pdu_format = (canID & 0xff0000) >> 16
        pdu_specific = (canID & 0xff00) >> 8
        source_address = canID & 0xff
        return klass(priority, reserved, datapage, pdu_format, pdu_specific, source_address)

    @property
    def pgn(self):
        return (self.pdu_format << 8) | self.pdu_specific

    @pgn.setter
    def pgn(self, value):
        """

        """
        self.pdu_format = (value & 0xff00) >> 8
        self.pdu_specific = (value & 0xff)


    @property
    def canID(self):
        """
        """
        return ((self.priority & 0x07) << 26) | ((self.reserved & 0x01) << 25) | ((self.datapage & 0x01) << 24) | \
            ((self.pdu_format & 0xff) << 16) |  ((self.pdu_specific & 0xff) << 8) | (self.source_address & 0xff)


    @canID.setter
    def canID(self, value):
        pass

    def __str__(self):
        return "{}(priority = {}, reserved = {}, datapage = {}, pdu_format = {}, pdu_specific = {}, source_address = {})".\
            format(self.__class__.__name__, self.priority, self.reserved, self.datapage, self.pdu_format,
            self.pdu_specific, self.source_address
        )

    __repr__ = __str__


class LinProductIdType(object):
    """

    """

    def __init__(self, supplier_id, function_id, variant = 0):
        self.supplier_id = supplier_id & 0xffff
        self.function_id = function_id & 0xffff
        self.variant = variant & 0xff

    def __str__(self):
        return "LinProductIdType(supplier_id = {}, function_id = {}, variant = {})".format(
            self.supplier_id, self.function_id, self.variant)

    __repr__ = __str__




from pydbc import types

def test_j1939_address():
    j0 = types.J1939Address.from_int(217056510)
    assert j0.priority == 3
    assert j0.reserved == 0
    assert j0.datapage == 0
    assert j0.pdu_format == 240
    assert j0.pdu_specific == 4
    assert j0.source_address == 254
    assert j0.pgn == 61444
    assert j0.canID == 217056510

def test_j1939_set_pgn():
    j0 = types.J1939Address.from_int(217056510)
    j0.pgn = 64850
    assert j0.canID == 217928446
    assert j0.pgn == 64850

def test_j1939_str(capsys):
    j0 = types.J1939Address.from_int(419283454)
    print(str(j0))
    captured = capsys.readouterr()
    assert captured.out == "J1939Address(priority = 6, reserved = 0, datapage = 0, pdu_format = 253, pdu_specific = 193, source_address = 254)\n"

def test_j1939_repr(capsys):
    j0 = types.J1939Address.from_int(419283454)
    print(repr(j0))
    captured = capsys.readouterr()
    assert captured.out == "J1939Address(priority = 6, reserved = 0, datapage = 0, pdu_format = 253, pdu_specific = 193, source_address = 254)\n"

def test_lin_product_id():
    lp = types.LinProductIdType(0x3e, 2, 4)
    assert lp.supplier_id == 0x3e
    assert lp.function_id == 2
    assert lp.variant == 4

def test_lin_product_id_str(capsys):
    lp = types.LinProductIdType(0x3e, 2, 4)
    print(str(lp))
    captured = capsys.readouterr()
    assert captured.out == "LinProductIdType(supplier_id = 62, function_id = 2, variant = 4)\n"

def test_lin_product_id_repr(capsys):
    lp = types.LinProductIdType(0x3e, 2, 4)
    print(repr(lp))
    captured = capsys.readouterr()
    assert captured.out == "LinProductIdType(supplier_id = 62, function_id = 2, variant = 4)\n"




#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.
   (C) 2010-2018 by Christoph Schueler <cpu12.gems@googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__  = 'Christoph Schueler'
__version__ = '0.9'

##
## Convenience functions for Mako Templates.
##

from io import StringIO
import re

from mako.template import Template
from mako.runtime import Context
from mako import exceptions


indentText = lambda text, leftmargin = 0: '\n'.join(["%s%s" % ((" " * leftmargin), line, ) for line in text.splitlines()])

def renderTemplate(filename, namespace = {}, leftMargin = 0, rightMargin = 80, formatExceptions = False, encoding = 'utf-8'):
    # TODO: filename und text parameter ==> Nur noch EINE funktion!!!
    buf = StringIO()
    ctx = Context(buf, **namespace)
    try:
        tobj = Template(filename = filename, output_encoding = encoding,  format_exceptions = formatExceptions)
        tobj.render_context(ctx)
    except:
        print(exceptions.text_error_template().render())
        return None
    ##return strings.reformat(buf.getvalue(), leftMargin, rightMargin)
    return buf.getvalue()

def renderTemplateFromText(tmpl, namespace = {}, leftMargin = 0, rightMargin = 80, formatExceptions = True, encoding = 'utf-8'):
    buf = StringIO()
    ctx = Context(buf, **namespace)
    try:
        tobj = Template(text = tmpl, output_encoding = encoding, format_exceptions = formatExceptions)
        tobj.render_context(ctx)
    except:
        print(exceptions.text_error_template().render())
        return None
    return indentText(buf.getvalue(), leftMargin) #, rightMargin)


def callDef(template, definition, *args, **kwargs):
    return template.get_def(definition).render(*args, **kwargs)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Vehicle Network Database Importer

This script imports vehicle network description files (.dbc, .ldf, .ncf) into a database.
It creates a .vndb file for each input file.
"""

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2010-2023 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__ = "Christoph Schueler"
__version__ = "0.1.0"

import argparse
import logging
import os
import pathlib
from typing import Optional, Union

from pydbc.parser import ParserWrapper
from pydbc.dbcListener import DbcListener

# from pydbc.ldfListener import LdfListener
# from pydbc.ncfListener import NcfListener
from pydbc.types import FileType


def parseFile(
    pth: pathlib.Path,
    filetype: FileType,
    debug: bool = False,
    remove_file: bool = False,
    logLevel: str = "WARN",
) -> Optional[object]:
    """Parse a vehicle network description file and create a database.

    Args:
        pth: Path to the file to parse
        filetype: Type of the file (DBC, LDF, NCF)
        debug: Enable debug output
        remove_file: Remove existing database file before parsing
        logLevel: Logging level (WARN, INFO, ERROR, DEBUG)

    Returns:
        SQLAlchemy session object or None if parsing failed
    """
    if filetype == FileType.DBC:
        grammar = "dbc"
        start_symbol = "dbcfile"
        listenerClass = DbcListener
    elif filetype == FileType.LDF:
        grammar = "ldf"
        start_symbol = "lin_description_file"
        listenerClass = LdfListener
    elif filetype == FileType.NCF:
        grammar = "ncf"
        start_symbol = "toplevel"
        listenerClass = NcfListener
    else:
        raise ValueError(f"Invalid filetype '{filetype}'")

    parser = ParserWrapper(
        grammar, start_symbol, listenerClass, debug=debug, logLevel=logLevel
    )
    logging.info(f"Processing '{pth}'")

    dbfn = f"{pth.stem}.vndb"
    if remove_file:
        try:
            os.unlink(dbfn)
        except FileNotFoundError:
            # File doesn't exist, no need to remove
            pass
        except PermissionError as e:
            logging.error(f"Permission error removing {dbfn}: {e}")
            return None
        except Exception as e:
            logging.error(f"Error removing {dbfn}: {e}")
            return None

    try:
        session = parser.parseFromFile(str(pth))
        logging.info(f"Successfully parsed {pth}")
        return session
    except Exception as e:
        logging.error(f"Error parsing {pth}: {e}")
        return None


def get_file_type(pth: pathlib.Path) -> Optional[FileType]:
    """Determine the file type based on the file extension.

    Args:
        pth: Path to the file

    Returns:
        FileType enum value or None if the file type is not supported
    """
    suffix = pth.suffix.lower()
    if suffix == ".dbc":
        result = FileType.DBC
    elif suffix == ".ldf":
        result = FileType.LDF
    elif suffix == ".ncf":
        result = FileType.NCF
    else:
        result = None
    return result


def importFile(pth: pathlib.Path, logLevel: str) -> Optional[object]:
    """Import a vehicle network description file into a database.

    Args:
        pth: Path to the file to import
        logLevel: Logging level (WARN, INFO, ERROR, DEBUG)

    Returns:
        SQLAlchemy session object or None if import failed
    """
    file_type = get_file_type(pth)
    if file_type is None:
        logging.error(f"Unsupported file type: {pth.suffix}")
        return None

    return parseFile(pth, file_type, remove_file=True, logLevel=logLevel)


def main() -> None:
    """Main entry point for the vndb_importer script."""
    footer = (
        "CAVEAT: In this version vndb_importer is DESTRUCTIVE, i.e. no merging happens!"
    )
    parser = argparse.ArgumentParser(
        description="Import vehicle network description files into a database.",
        epilog=footer,
    )
    parser.add_argument(
        "vehicle_file",
        help=".dbc, .ldf, or .ncf file(s) to import (glob patterns supported)",
        nargs="+",
    )
    parser.add_argument(
        "-k",
        dest="keepDirectory",
        action="store_true",
        default=False,
        help="keep directory; otherwise create db in current directory",
    )
    parser.add_argument(
        "-l",
        dest="loglevel",
        type=str,
        choices=["debug", "info", "warn", "error"],
        default="warn",
        help="logging level [debug | info | warn | error]",
    )
    parser.add_argument(
        "-w",
        dest="winout",
        action="store_true",
        help="Format output for Windows console.",
    )
    parser.add_argument(
        "-u",
        dest="ucout",
        action="store_true",
        help="Generate UTF-8 encoded output (otherwise Latin-1).",
    )

    args = parser.parse_args()

    # Configure logging
    log_level = getattr(logging, args.loglevel.upper())
    logging.basicConfig(level=log_level, format="%(levelname)s: %(message)s")

    # Process each input file
    for arg in args.vehicle_file:
        # print(arg)
        pth=pathlib.Path(arg)
        if not pth.exists():
            logging.error(f"File not found: {pth}")
            continue

        if not pth.is_file():
            logging.error(f"Not a file: {pth}")
            continue

        # Import the file
        session = importFile(pth, args.loglevel)
        if session:
            logging.info(f"Successfully imported {pth}")
        else:
            logging.error(f"Failed to import {pth}")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Creational API for NCF (Network Configuration File) components.

This module provides a high-level API for creating and manipulating NCF components
such as vehicles, ECUs, networks, and gateways.
"""

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2010-2023 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__ = "Christoph Schueler"
__version__ = "0.1.0"

from typing import Optional, Union

from pydbc.db import VNDB
from pydbc.db.model import (
    Network,
    Node,
    Signal,
    ECU,
    Vehicle,
    Gateway_Signal,
    EnvVar,
    EnvVar_AccessNode,
    ECU_EnvVar,
    ECU_Node,
    Vehicle_ECU,
    Vehicle_Network,
    Network_Node,
    EnvironmentVariablesData,
)


class NCFCreator:
    """High-level API for creating NCF components."""

    def __init__(self, db_path: str = ":memory:", debug: bool = False):
        """Initialize a new NCF creator.

        Args:
            db_path: Path to the database file or ":memory:" for in-memory database
            debug: Enable debug mode
        """
        self.db = VNDB.create(db_path, debug=debug)
        self.session = self.db.session
        self._networks = {}
        self._nodes = {}
        self._signals = {}
        self._ecus = {}
        self._vehicles = {}
        self._env_vars = {}

    def create_network(self, name: str, **kwargs) -> Network:
        """Create a new network.

        Args:
            name: Name of the network
            **kwargs: Additional network attributes

        Returns:
            The created Network object
        """
        network = Network(name=name, **kwargs)
        self.session.add(network)
        self._networks[name] = network
        return network

    def create_node(self, name: str, **kwargs) -> Node:
        """Create a new node.

        Args:
            name: Name of the node
            **kwargs: Additional node attributes

        Returns:
            The created Node object
        """
        node = Node(name=name, **kwargs)
        self.session.add(node)
        self._nodes[name] = node
        return node

    def create_ecu(self, name: str, **kwargs) -> ECU:
        """Create a new ECU.

        Args:
            name: Name of the ECU
            **kwargs: Additional ECU attributes

        Returns:
            The created ECU object
        """
        ecu = ECU(name=name, **kwargs)
        self.session.add(ecu)
        self._ecus[name] = ecu
        return ecu

    def create_vehicle(self, name: str, **kwargs) -> Vehicle:
        """Create a new vehicle.

        Args:
            name: Name of the vehicle
            **kwargs: Additional vehicle attributes

        Returns:
            The created Vehicle object
        """
        vehicle = Vehicle(name=name, **kwargs)
        self.session.add(vehicle)
        self._vehicles[name] = vehicle
        return vehicle

    def create_env_var(
        self,
        name: str,
        var_type: str,
        unit: Optional[str] = None,
        minimum: Optional[float] = None,
        maximum: Optional[float] = None,
        initial_value: Optional[str] = None,
        access_type: Optional[str] = None,
        access_node: Optional[int] = None,
        **kwargs
    ) -> EnvVar:
        """Create a new environment variable.

        Args:
            name: Name of the environment variable
            var_type: Type of the environment variable
            unit: Unit of the environment variable
            minimum: Minimum value
            maximum: Maximum value
            initial_value: Initial value
            access_type: Access type
            access_node: Access node ID
            **kwargs: Additional environment variable attributes

        Returns:
            The created EnvVar object
        """
        env_var = EnvVar(
            name=name,
            var_type=var_type,
            unit=unit,
            minimum=minimum,
            maximum=maximum,
            initial_value=initial_value,
            access_type=access_type,
            access_node=access_node,
            **kwargs
        )
        self.session.add(env_var)
        self._env_vars[name] = env_var
        return env_var

    def create_gateway_signal(
        self,
        source_signal: Union[str, Signal, int],
        target_signal: Union[str, Signal, int],
        **kwargs
    ) -> Gateway_Signal:
        """Create a new gateway signal mapping.

        Args:
            source_signal: Source signal name, Signal object, or rid
            target_signal: Target signal name, Signal object, or rid
            **kwargs: Additional gateway signal attributes

        Returns:
            The created Gateway_Signal object
        """
        if isinstance(source_signal, str):
            source_signal = (
                self.session.query(Signal).filter_by(name=source_signal).first().rid
            )
        elif isinstance(source_signal, Signal):
            source_signal = source_signal.rid

        if isinstance(target_signal, str):
            target_signal = (
                self.session.query(Signal).filter_by(name=target_signal).first().rid
            )
        elif isinstance(target_signal, Signal):
            target_signal = target_signal.rid

        gateway_signal = Gateway_Signal(
            source_signal=source_signal, target_signal=target_signal, **kwargs
        )
        self.session.add(gateway_signal)
        return gateway_signal

    def add_node_to_network(
        self,
        network: Union[str, Network],
        node: Union[str, Node],
        connector_name: Optional[str] = None,
    ) -> Network_Node:
        """Add a node to a network.

        Args:
            network: Network name or Network object
            node: Node name or Node object
            connector_name: Name of the connector

        Returns:
            The created Network_Node object
        """
        if isinstance(network, str):
            network = self._networks[network]
        if isinstance(node, str):
            node = self._nodes[node]

        network_node = Network_Node(
            network=network, node=node, connector_name=connector_name
        )
        self.session.add(network_node)
        return network_node

    def add_ecu_to_vehicle(
        self, vehicle: Union[str, Vehicle], ecu: Union[str, ECU]
    ) -> Vehicle_ECU:
        """Add an ECU to a vehicle.

        Args:
            vehicle: Vehicle name or Vehicle object
            ecu: ECU name or ECU object

        Returns:
            The created Vehicle_ECU object
        """
        if isinstance(vehicle, str):
            vehicle = self._vehicles[vehicle]
        if isinstance(ecu, str):
            ecu = self._ecus[ecu]

        vehicle_ecu = Vehicle_ECU(vehicle=vehicle, ecu=ecu)
        self.session.add(vehicle_ecu)
        return vehicle_ecu

    def add_network_to_vehicle(
        self, vehicle: Union[str, Vehicle], network: Union[str, Network]
    ) -> Vehicle_Network:
        """Add a network to a vehicle.

        Args:
            vehicle: Vehicle name or Vehicle object
            network: Network name or Network object

        Returns:
            The created Vehicle_Network object
        """
        if isinstance(vehicle, str):
            vehicle = self._vehicles[vehicle]
        if isinstance(network, str):
            network = self._networks[network]

        vehicle_network = Vehicle_Network(vehicle=vehicle, network=network)
        self.session.add(vehicle_network)
        return vehicle_network

    def add_node_to_ecu(self, ecu: Union[str, ECU], node: Union[str, Node]) -> ECU_Node:
        """Add a node to an ECU.

        Args:
            ecu: ECU name or ECU object
            node: Node name or Node object

        Returns:
            The created ECU_Node object
        """
        if isinstance(ecu, str):
            ecu = self._ecus[ecu]
        if isinstance(node, str):
            node = self._nodes[node]

        ecu_node = ECU_Node(ecu=ecu, node=node)
        self.session.add(ecu_node)
        return ecu_node

    def add_env_var_to_ecu(
        self, ecu: Union[str, ECU], env_var: Union[str, EnvVar]
    ) -> ECU_EnvVar:
        """Add an environment variable to an ECU.

        Args:
            ecu: ECU name or ECU object
            env_var: Environment variable name or EnvVar object

        Returns:
            The created ECU_EnvVar object
        """
        if isinstance(ecu, str):
            ecu = self._ecus[ecu]
        if isinstance(env_var, str):
            env_var = self._env_vars[env_var]

        ecu_env_var = ECU_EnvVar(ecu=ecu, env_var=env_var)
        self.session.add(ecu_env_var)
        return ecu_env_var

    def add_access_node_to_env_var(
        self, env_var: Union[str, EnvVar], node: Union[str, Node], access_type: str
    ) -> EnvVar_AccessNode:
        """Add an access node to an environment variable.

        Args:
            env_var: Environment variable name or EnvVar object
            node: Node name or Node object
            access_type: Access type (e.g., "read", "write", "readWrite")

        Returns:
            The created EnvVar_AccessNode object
        """
        if isinstance(env_var, str):
            env_var = self._env_vars[env_var]
        if isinstance(node, str):
            node = self._nodes[node]

        env_var_access_node = EnvVar_AccessNode(
            env_var=env_var, node=node, access_type=access_type
        )
        self.session.add(env_var_access_node)
        return env_var_access_node

    def create_env_var_data(
        self, env_var: Union[str, EnvVar], data: str
    ) -> EnvironmentVariablesData:
        """Create environment variable data.

        Args:
            env_var: Environment variable name or EnvVar object
            data: Data value

        Returns:
            The created EnvironmentVariablesData object
        """
        if isinstance(env_var, str):
            env_var = self._env_vars[env_var]

        env_var_data = EnvironmentVariablesData(env_var=env_var, data=data)
        self.session.add(env_var_data)
        return env_var_data

    def commit(self):
        """Commit changes to the database."""
        self.session.commit()

    def close(self):
        """Close the database connection."""
        self.session.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Creational API for DBC (CAN Database) components.

This module provides a high-level API for creating and manipulating DBC components
such as networks, nodes, messages, and signals.
"""

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2010-2023 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__ = "Christoph Schueler"
__version__ = "0.1.0"

from typing import Optional, Union, Dict, Any

from pydbc.db import VNDB
from pydbc.db.model import (
    Network,
    Node,
    Message,
    Signal,
    Message_Signal,
    Value_Description,
    Valuetable,
    Node_RxSignal,
    Attribute_Definition,
    Attribute_Value,
)


class DBCCreator:
    """High-level API for creating DBC components.

    This creator can either create a new database (default) or attach to an existing
    SQLAlchemy session returned by the parser. Use DBCCreator.from_session(session)
    or pass session=... to the constructor to extend an existing database.
    """

    def __init__(self, db_path: str = ":memory:", debug: bool = False, session=None):
        """Initialize a new DBC creator.

        Args:
            db_path: Path to the database file or ":memory:" for in-memory database
            debug: Enable debug mode
            session: Optional existing SQLAlchemy session to attach to (e.g., from ParserWrapper)
        """
        if session is not None:
            # Attach to an existing session (parsed DB, etc.)
            self.db = None  # no VNDB instance managed here
            self.session = session
        else:
            self.db = VNDB.create(db_path, debug=debug, autocommit=True)
            self.session = self.db.session
        self._networks = {}
        self._nodes = {}
        self._messages = {}
        self._signals = {}
        self._valuetables = {}

    @classmethod
    def from_session(cls, session):
        """Create a DBCCreator attached to an existing SQLAlchemy session.

        Example:
            from pydbc.parser import ParserWrapper
            from pydbc.dbcListener import DbcListener
            wrapper = ParserWrapper(grammarName="dbc", startSymbol="dbcfile", listenerClass=DbcListener)
            session = wrapper.parseFromFile("C:\\path\\to\\file.dbc")
            dbc = DBCCreator.from_session(session)
        """
        return cls(session=session)

    # internal helpers for lazy name -> object resolution
    def _get_node(self, node):
        if isinstance(node, str):
            if node in self._nodes:
                return self._nodes[node]
            obj = self.session.query(Node).filter_by(name=node).first()
            if obj is None:
                raise KeyError(f"Node '{node}' not found")
            self._nodes[node] = obj
            return obj
        return node

    def _get_message(self, message):
        if isinstance(message, str):
            if message in self._messages:
                return self._messages[message]
            obj = self.session.query(Message).filter_by(name=message).first()
            if obj is None:
                raise KeyError(f"Message '{message}' not found")
            self._messages[message] = obj
            return obj
        return message

    def _get_signal(self, signal):
        if isinstance(signal, str):
            if signal in self._signals:
                return self._signals[signal]
            obj = self.session.query(Signal).filter_by(name=signal).first()
            if obj is None:
                raise KeyError(f"Signal '{signal}' not found")
            self._signals[signal] = obj
            return obj
        return signal

    def index_existing(self):
        """Optionally preload existing objects from the current session into caches."""
        for n in self.session.query(Network).all():
            self._networks[n.name] = n
        for n in self.session.query(Node).all():
            self._nodes[n.name] = n
        for m in self.session.query(Message).all():
            if m.name:
                self._messages[m.name] = m
        for s in self.session.query(Signal).all():
            self._signals[s.name] = s

    def create_network(self, name: str, **kwargs) -> Network:
        """Create a new network.

        Args:
            name: Name of the network
            **kwargs: Additional network attributes

        Returns:
            The created Network object
        """
        network = Network(name=name, **kwargs)
        self.session.add(network)
        self._networks[name] = network
        return network

    def create_node(self, name: str, **kwargs) -> Node:
        """Create a new node.

        Args:
            name: Name of the node
            **kwargs: Additional node attributes

        Returns:
            The created Node object
        """
        node = Node(name=name, **kwargs)
        self.session.add(node)
        self._nodes[name] = node
        return node

    def create_message(
        self,
        name: str,
        message_id: int,
        dlc: int,
        sender: Union[str, Node, int],
        **kwargs
    ) -> Message:
        """Create a new message.

        Args:
            name: Name of the message
            message_id: Message ID
            dlc: Data Length Code (message size in bytes)
            sender: Sender node (name, Node object, or rid)
            **kwargs: Additional message attributes

        Returns:
            The created Message object
        """
        if isinstance(sender, str):
            sender_obj = self._get_node(sender)
            sender_rid = sender_obj.rid
        elif isinstance(sender, Node):
            sender_rid = sender.rid
        elif isinstance(sender, int):
            sender_rid = sender
        else:
            raise TypeError("sender must be a node name, Node, or integer rid")

        message = Message(name=name, message_id=message_id, dlc=dlc, sender=sender_rid, **kwargs)
        self.session.add(message)
        if name:
            self._messages[name] = message
        return message

    def create_signal(
        self,
        name: str,
        bitsize: int,
        byteorder: int = 1,
        sign: int = 1,
        formula_factor: float = 1.0,
        formula_offset: float = 0.0,
        minimum: float = 0.0,
        maximum: float = 0.0,
        unit: str = "",
        **kwargs
    ) -> Signal:
        """Create a new signal.

        Args:
            name: Name of the signal
            bitsize: Size of the signal in bits
            byteorder: Byte order (1 for little endian, 0 for big endian)
            sign: Sign (1 for unsigned, 0 for signed)
            formula_factor: Factor for physical value calculation
            formula_offset: Offset for physical value calculation
            minimum: Minimum value
            maximum: Maximum value
            unit: Unit of the signal
            **kwargs: Additional signal attributes

        Returns:
            The created Signal object
        """
        signal = Signal(
            name=name,
            bitsize=bitsize,
            byteorder=byteorder,
            sign=sign,
            formula_factor=formula_factor,
            formula_offset=formula_offset,
            minimum=minimum,
            maximum=maximum,
            unit=unit,
            **kwargs
        )
        self.session.add(signal)
        self._signals[name] = signal
        return signal

    def add_signal_to_message(
        self,
        message: Union[str, Message],
        signal: Union[str, Signal],
        offset: int,
        multiplexor_signal: Optional[int] = None,
        multiplex_dependent: Optional[bool] = None,
        multiplexor_value: Optional[int] = None,
    ) -> Message_Signal:
        """Add a signal to a message.

        Args:
            message: Message name or Message object
            signal: Signal name or Signal object
            offset: Bit offset in the message
            multiplexor_signal: Multiplexor signal ID
            multiplex_dependent: Whether the signal is multiplex dependent
            multiplexor_value: Multiplexor value

        Returns:
            The created Message_Signal object
        """
        if isinstance(message, str):
            message = self._get_message(message)
        if isinstance(signal, str):
            signal = self._get_signal(signal)

        msg_sig = Message_Signal(
            message=message,
            signal=signal,
            offset=offset,
            multiplexor_signal=multiplexor_signal,
            multiplex_dependent=multiplex_dependent,
            multiplexor_value=multiplexor_value,
        )
        self.session.add(msg_sig)
        return msg_sig

    def create_valuetable(self, name: str, values: Dict[int, str]) -> Valuetable:
        """Create a value table for signal value descriptions.

        Args:
            name: Name of the value table
            values: Dictionary mapping values to descriptions

        Returns:
            The created Valuetable object
        """
        value_descriptions = []
        for value, description in values.items():
            vd = Value_Description(value=value, value_description=description)
            value_descriptions.append(vd)
            self.session.add(vd)

        valuetable = Valuetable(name=name, values=value_descriptions)
        self.session.add(valuetable)
        self._valuetables[name] = valuetable
        return valuetable

    def add_node_as_receiver(
        self, signal: Union[str, Signal], node: Union[str, Node]
    ) -> Node_RxSignal:
        """Add a node as a receiver for a signal.

        Args:
            signal: Signal name or Signal object
            node: Node name or Node object

        Returns:
            The created Node_RxSignal object
        """
        if isinstance(signal, str):
            signal = self._get_signal(signal)
        if isinstance(node, str):
            node = self._get_node(node)

        node_rx_signal = Node_RxSignal(signal=signal, node=node)
        self.session.add(node_rx_signal)
        return node_rx_signal

    def create_attribute_definition(
        self, name: str, object_type: str, value_type: str, **kwargs
    ) -> Attribute_Definition:
        """Create an attribute definition.

        Args:
            name: Name of the attribute
            object_type: Type of object the attribute applies to
            value_type: Type of the attribute value
            **kwargs: Additional attribute definition properties

        Returns:
            The created Attribute_Definition object
        """
        attr_def = Attribute_Definition(
            name=name, object_type=object_type, value_type=value_type, **kwargs
        )
        self.session.add(attr_def)
        return attr_def

    def set_attribute_value(
        self,
        attribute_definition: Union[str, Attribute_Definition],
        object_id: int,
        value: Any,
    ) -> Attribute_Value:
        """Set an attribute value for an object.

        Args:
            attribute_definition: Attribute definition name or object
            object_id: ID of the object to set the attribute for
            value: Value to set

        Returns:
            The created Attribute_Value object
        """
        if isinstance(attribute_definition, str):
            attribute_definition = (
                self.session.query(Attribute_Definition)
                .filter_by(name=attribute_definition)
                .first()
            )

        attr_value = Attribute_Value(
            attribute_definition=attribute_definition,
            object_id=object_id,
            value=str(value),
        )
        self.session.add(attr_value)
        return attr_value

    def commit(self):
        """Commit changes to the database."""
        self.session.commit()

    def close(self):
        """Close the database connection."""
        self.session.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

__copyright__ = """
   pySART - Simplified AUTOSAR-Toolkit for Python.

   (C) 2010-2021 by Christoph Schueler <cpu12.gems.googlemail.com>

   All Rights Reserved

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

   s. FLOSS-EXCEPTION.txt
"""
__author__ = "Christoph Schueler"
__version__ = "0.2.0"  # Updated to Python 3.10 and SQLAlchemy 2.0

import datetime
import re
from typing import Optional, List, Union

from sqlalchemy import BLOB, DateTime, String, Integer, Float, Boolean, Unicode
from sqlalchemy import Column, ForeignKey, ForeignKeyConstraint, UniqueConstraint, func
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.orderinglist import ordering_list
from sqlalchemy.orm import (
    DeclarativeBase,
    declared_attr,
    relationship,
    Mapped,
    mapped_column,
)

from pydbc.types import LinProductIdType, ByteOrderType, ValueType


class Base(DeclarativeBase):
    """Base class for all models"""

    pass


# Regular expression to match initial values (either array of integers or a single integer)
INIT_VALUE = re.compile(r"(?P<array>\d+(?:,\d+)+) | (?P<scalar>\d+)", re.VERBOSE)


class MixInBase:
    """Base mixin for all models"""

    @declared_attr.directive
    def __tablename__(cls) -> str:
        return cls.__name__.lower()

    def __repr__(self) -> str:
        columns = [c.name for c in self.__class__.__table__.c]
        result = []
        for name, value in [(n, getattr(self, n)) for n in columns]:
            if isinstance(value, str):
                result.append(f"{name} = '{value}'")
            else:
                result.append(f"{name} = {value}")
        return f"{self.__class__.__name__}({', '.join(result)})"


class CommentableMixIn:
    """Mixin for models with comments"""

    comment: Mapped[Optional[str]] = mapped_column(Unicode(255), default=None)


class MinMaxMixIn:
    """Mixin for models with min/max values"""

    minimum: Mapped[float] = mapped_column(Float, default=0.0)
    maximum: Mapped[float] = mapped_column(Float, default=0.0)


class RidMixIn(MixInBase):
    """Mixin for models with a primary key 'rid'"""

    rid: Mapped[int] = mapped_column("rid", Integer, primary_key=True)


def StdInteger(
    default: int = 0,
    primary_key: bool = False,
    unique: bool = False,
    nullable: bool = False,
) -> mapped_column:
    """Helper function to create a standard integer column using SQLAlchemy 2.0 style"""
    return mapped_column(
        Integer,
        default=default,
        nullable=nullable,
        primary_key=primary_key,
        unique=unique,
    )


def StdFloat(
    default: float = 0.0,
    primary_key: bool = False,
    unique: bool = False,
    nullable: bool = False,
) -> mapped_column:
    """Helper function to create a standard float column using SQLAlchemy 2.0 style"""
    return mapped_column(
        Float,
        default=default,
        nullable=nullable,
        primary_key=primary_key,
        unique=unique,
    )


class Node(Base, RidMixIn, CommentableMixIn):
    """Node model representing a network node"""

    name: Mapped[str] = mapped_column(
        Unicode(255), nullable=False, unique=True, index=True
    )
    node_id: Mapped[int] = mapped_column(Integer, default=0)
    type_: Mapped[str] = mapped_column(String(256))

    __mapper_args__ = {
        "polymorphic_identity": "Node",
        "polymorphic_on": type_,
    }


class Message_Signal(Base, MixInBase):
    """Association model between Message and Signal"""

    message_id: Mapped[int] = mapped_column(
        Integer,
        ForeignKey("message.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    signal_id: Mapped[int] = mapped_column(
        Integer,
        ForeignKey("signal.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    offset: Mapped[int] = mapped_column(Integer, nullable=False, default=0)
    multiplexor_signal: Mapped[Optional[int]] = mapped_column(Integer, default=0)
    multiplex_dependent: Mapped[Optional[bool]] = mapped_column(Boolean)
    multiplexor_value: Mapped[Optional[int]] = mapped_column(Integer)

    signal: Mapped["Signal"] = relationship("Signal", lazy="joined")
    message: Mapped["Message"] = relationship(
        "Message", back_populates="message_signals"
    )

    def __init__(
        self,
        signal: "Signal",
        message: "Message",
        offset: int,
        multiplexor_signal: Optional[int] = None,
        multiplex_dependent: Optional[bool] = None,
        multiplexor_value: Optional[int] = None,
    ):
        self.signal = signal
        self.message = message
        self.offset = offset
        self.multiplexor_signal = multiplexor_signal
        self.multiplex_dependent = multiplex_dependent
        self.multiplexor_value = multiplexor_value


class Message(Base, RidMixIn, CommentableMixIn):
    """Message model representing a network message"""

    name: Mapped[Optional[str]] = mapped_column(Unicode(255), nullable=True, index=True)
    message_id: Mapped[Optional[int]] = mapped_column(
        Integer, nullable=True, default=None
    )
    dlc: Mapped[int] = mapped_column(Integer, default=0)
    sender: Mapped[int] = mapped_column(Integer, default=0)

    # transmitter_node_id

    type: Mapped[str] = mapped_column(String(256))

    # Relationships
    message_signals: Mapped[List["Message_Signal"]] = relationship(
        "Message_Signal", cascade="all, delete-orphan", back_populates="message"
    )
    signals = association_proxy("message_signals", "signal")

    __mapper_args__ = {
        "polymorphic_identity": "Message",
        "polymorphic_on": type,
    }


class Signal(Base, RidMixIn, CommentableMixIn):
    """Signal model representing a network signal"""

    name: Mapped[str] = mapped_column(Unicode(255), nullable=False, index=True)
    bitsize: Mapped[int] = mapped_column(Integer, default=0)
    byteorder: Mapped[int] = mapped_column(Integer, default=1)
    sign: Mapped[int] = mapped_column(Integer, default=1)
    valuetype: Mapped[int] = mapped_column(Integer, default=0)
    formula_factor: Mapped[float] = mapped_column(Float, default=1.0)
    formula_offset: Mapped[float] = mapped_column(Float, default=0.0)
    minimum: Mapped[float] = mapped_column(Float, default=0.0)
    maximum: Mapped[float] = mapped_column(Float, default=0.0)
    unit: Mapped[Optional[str]] = mapped_column(Unicode(255))
    type: Mapped[str] = mapped_column(String(256))

    __mapper_args__ = {
        "polymorphic_identity": "Signal",
        "polymorphic_on": type,
    }


class Network(Base, RidMixIn, CommentableMixIn):
    """Network model representing a communication network"""

    name: Mapped[str] = mapped_column(
        Unicode(255), nullable=False, unique=True, index=True
    )
    protocol: Mapped[int] = mapped_column(Integer, default=0)
    baudrate: Mapped[int] = mapped_column(Integer, default=0)
    type: Mapped[str] = mapped_column(String(256))

    __mapper_args__ = {
        "polymorphic_identity": "Network",
        "polymorphic_on": type,
    }


class Node_RxSig(Base, MixInBase):
    node = Column(
        Integer,
        ForeignKey("node.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    signal = Column(
        Integer,
        ForeignKey("signal.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )


class Node_RxSignal(Base, MixInBase):
    node_id = Column(
        Integer,
        ForeignKey("node.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    message_id = Column(
        Integer,
        ForeignKey("message.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    signal_id = Column(
        Integer,
        ForeignKey("signal.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    __table_args__ = (
        ForeignKeyConstraint(
            columns=["message_id", "signal_id"],
            refcolumns=["message_signal.message_id", "message_signal.signal_id"],
        ),
    )

    # receiver_id = relationship(
    #    "Node_RxSignal", cascade="all, delete-orphan", backref="signal"
    # )
    # receiver = association_proxy("receiver_id", "signal")

    node = relationship("Node", uselist=False, lazy="joined")
    message = relationship("Message", uselist=False, lazy="joined")
    signal = relationship(
        "Signal", uselist=False, lazy="joined"
    )  # backref = "rx_signals", uselist  = False,


class Node_TxMessage(Base, MixInBase):
    node_id = Column(
        Integer,
        ForeignKey("node.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    message_id = Column(
        Integer,
        ForeignKey("message.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    node = relationship("Node", backref="tx_messages", uselist=False)
    message = relationship("Message", backref="tx_messages", uselist=False)


class Node_TxSig(Base, MixInBase):
    node_id = Column(
        Integer,
        ForeignKey("node.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    signal_id = Column(
        Integer,
        ForeignKey("signal.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    node = relationship("Node", backref="tx_sigs", uselist=False)
    signal = relationship("Signal", backref="tx_sigs", uselist=False)


class Attribute_Definition(Base, RidMixIn, CommentableMixIn):
    name = Column(Unicode(255), nullable=False, unique=True, index=True)
    objecttype = StdInteger()
    valuetype = StdInteger()
    array = Column(Boolean, default=False)
    minimum = Column(Float, default=0.0)
    maximum = Column(Float, default=0.0)
    enumvalues = Column(Unicode(255))
    default_number = Column(Float, default=0.0)
    default_string = Column(Unicode(255), default=None)


class Attribute_Value(Base, MixInBase):
    object_id = Column(Integer, nullable=False, default=0, primary_key=True)
    attribute_definition_id = Column(
        Integer,
        ForeignKey("attribute_definition.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    num_value = Column(Float, default=0.0)
    string_value = Column(Unicode(255))
    attribute_definition = relationship("Attribute_Definition", backref="values")

    """
    @hybrid_property
    def value(self):
        return self.num_value

    @value.setter
    def value(self, v):
        print("Setting value to '{}'".format(v))

    @value.expression
    def value(self):
        return self.num_value
    """


class AttributeRel_Value(Base, MixInBase):
    object_id = Column(Integer, nullable=False, default=0, primary_key=True)
    attribute_definition_id = Column(
        Integer,
        ForeignKey("attribute_definition.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    num_value = Column(Float, default=0.0)
    string_value = Column(Unicode(255))
    opt_object_id_1 = Column(Integer, nullable=False, default=0, primary_key=True)
    opt_object_id_2 = Column(Integer, nullable=False, default=0, primary_key=True)
    blob_value = Column(BLOB)
    attribute_definition = relationship(
        "Attribute_Definition", backref="attribute_rel_values"
    )


class VndbMeta(Base, RidMixIn):
    schema_version = StdInteger()
    vndb_type = StdInteger()
    created = Column(DateTime, default=datetime.datetime.now)


class Vndb_Migrations(Base, RidMixIn):
    app = Column(Unicode(255), nullable=False)
    name = Column(Unicode(255), nullable=False)
    applied = Column(DateTime, func.now)


class Vndb_Protocol(Base, MixInBase):
    network_id = Column(
        Integer,
        ForeignKey("network.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    name = Column(Unicode(255), nullable=False)
    specific = Column(Unicode(255), nullable=True)
    network = relationship("Network", uselist=False)


class ECU(Base, RidMixIn, CommentableMixIn):
    name = Column(Unicode(255), nullable=False, unique=True, index=True)


class ECU_EnvVar(Base, MixInBase):
    ecu = Column(
        Integer,
        ForeignKey("ecu.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    env_var = Column(
        Integer,
        ForeignKey("envvar.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )


class ECU_Node(Base, MixInBase):
    ecu = Column(
        Integer,
        ForeignKey("ecu.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    node = Column(
        Integer,
        ForeignKey("node.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )


class Gateway_Signal(Base, RidMixIn, CommentableMixIn):
    vehicle_id = Column(Integer, nullable=False, default=0)
    dest_signal = Column(Integer, nullable=False, default=0)
    dest_network = Column(Integer, nullable=False, default=0)
    dest_message = Column(Integer, nullable=False, default=0)
    dest_transmitter = Column(Integer, nullable=False, default=0)
    gateway = Column(Integer, nullable=False, default=0)
    source_signal = Column(Integer, nullable=False, default=0)
    source_network = Column(Integer, nullable=False, default=0)
    source_message = Column(Integer, nullable=False, default=0)
    source_receiver = Column(Integer, nullable=False, default=0)
    reserved_id1 = Column(Integer, nullable=False, default=0)


class Node_Group(Base, RidMixIn, CommentableMixIn):
    name = Column(Unicode(255), nullable=False, unique=True, index=True)
    node_object_type = Column(Integer, nullable=False, default=0)
    node_group_type = Column(Integer, nullable=False, default=0)


class Node_Group_Object(Base, MixInBase):
    parent_type = Column(Integer, nullable=False, default=0, primary_key=True)
    parent_rid = Column(Integer, nullable=False, default=0, primary_key=True)
    object_type = Column(Integer, nullable=False, default=0, primary_key=True)
    object_rid = Column(Integer, nullable=False, default=0, primary_key=True)
    object_rid2 = Column(Integer, nullable=False, default=0, primary_key=True)
    opt_object_ref_1 = Column(Integer, nullable=False, default=0)
    opt_object_ref_2 = Column(Integer, nullable=False, default=0)
    opt_object_value = Column(Integer, nullable=False, default=0)


class Network_Node(Base, MixInBase):
    network = Column(
        Integer,
        ForeignKey("network.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    node = Column(
        Integer,
        ForeignKey("node.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )


class Signal_Group(Base, RidMixIn):
    name = Column(Unicode(255), nullable=False)
    value = Column(Integer, nullable=False, default=0)
    message_id = Column(
        Integer,
        ForeignKey("message.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
    )
    message = relationship("Message", backref="signal_groups")
    UniqueConstraint("Message", "Name")


class Signal_Group_Signal(Base, MixInBase):
    signal_group_id = Column(
        Integer,
        ForeignKey("signal_group.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    message_id = Column(
        Integer,
        ForeignKey("message.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    signal_id = Column(
        Integer,
        ForeignKey("signal.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    signal_group = relationship("Signal_Group", backref="signal_group_signal")
    message = relationship("Message", backref="signal_group_signal")
    signal = relationship("Signal", backref="signal_group_signal")


class Valuetable(Base, RidMixIn, CommentableMixIn):
    """ """

    name = Column(Unicode(255), nullable=False, index=True)


class Object_Valuetable(Base, MixInBase):
    """ """

    object_type = Column(Integer, nullable=False, default=0)  # , primary_key = True
    object_rid = Column(Integer, nullable=False, default=0)  # , primary_key = True
    valuetable_id = Column(
        Integer,
        ForeignKey("valuetable.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    valuetable = relationship("Valuetable", backref="object_valuetable", uselist=False)


class Value_Description(Base, MixInBase):
    valuetable_id = Column(
        Integer,
        ForeignKey("valuetable.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    value = Column(Float, nullable=False, default=0.0, primary_key=True)
    value_description = Column(Unicode(255), nullable=False)
    valuetable = relationship("Valuetable", backref="values")


class Vehicle(Base, RidMixIn, CommentableMixIn):
    name = Column(Unicode(255), nullable=False, unique=True, index=True)


class Vehicle_ECU(Base, MixInBase):
    vehicle = Column(
        Integer,
        ForeignKey("vehicle.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )

    ecu = Column(
        Integer,
        ForeignKey("ecu.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )


class Vehicle_Network(Base, MixInBase):
    vehicle = Column(
        Integer,
        ForeignKey("vehicle.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )

    network = Column(
        Integer,
        ForeignKey("network.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )


class Versioninfo(Base, MixInBase):
    object_type = Column(Integer, nullable=False, default=0, primary_key=True)
    object_rid = Column(Integer, nullable=False, default=0, primary_key=True)
    version_number = Column(Integer, nullable=False, default=0)
    is_modified = Column(Boolean, nullable=False, default=False)


class Category_Definition(Base, RidMixIn):
    name = Column(Unicode(255), nullable=False)
    key = Column(Integer, nullable=False, default=0, unique=True)
    level = Column(Integer, nullable=False, default=0)


class Category_Value(Base, MixInBase):
    object_id = Column(Integer, nullable=False, default=0, primary_key=True)
    category_definition_id = Column(
        Integer,
        ForeignKey("category_definition.key", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
    )
    objecttype = Column(Integer, nullable=False, default=0, primary_key=True)
    category_definition = relationship("Category_Definition", backref="values")


class Dbc_Version(Base, RidMixIn):
    version_string = Column(Unicode(255), nullable=False, default="")
    network = Column(Integer, nullable=False, default=0, unique=True)


class EnvVar(Base, RidMixIn, CommentableMixIn):
    name = Column(Unicode(255), nullable=False, unique=True, index=True)
    type = StdInteger()
    unit = Column(Unicode(255), default=None)
    minimum = Column(Float, default=0.0)
    maximum = Column(Float, default=0.0)
    startup_value = Column(Float, default=0.0)
    size = StdInteger()
    access = StdInteger()
    accessingNodes = relationship("Node", secondary="envvar_accessnode")


class EnvVar_AccessNode(Base, MixInBase):
    envvar_id = Column(
        Integer,
        ForeignKey("envvar.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    node_id = Column(
        Integer,
        ForeignKey("node.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    # envvar = relationship("EnvVar", backref = "nodes")
    # node = relationship("Node", backref = "envvars")


class EnvironmentVariablesData(Base, MixInBase):
    name = Column(Unicode(255), nullable=False, primary_key=True)
    value = Column(Integer, nullable=False)


##
##  LIN specific classses.
##
class LinNetwork(Network):
    """LIN (Local Interconnect Network) network model"""

    lin_network_id: Mapped[int] = mapped_column(
        Integer, ForeignKey("network.rid"), primary_key=True
    )
    __mapper_args__ = {"polymorphic_identity": "LinNetwork"}

    protocol_version: Mapped[Optional[str]] = mapped_column(
        Unicode(255), nullable=True, default=None
    )
    language_version: Mapped[Optional[str]] = mapped_column(
        Unicode(255), nullable=True, default=None
    )
    file_revision: Mapped[Optional[str]] = mapped_column(
        Unicode(255), nullable=True, default=None
    )
    speed: Mapped[Optional[float]] = mapped_column(Float, nullable=True, default=None)
    channel_name: Mapped[Optional[str]] = mapped_column(
        Unicode(255), nullable=True, default=None
    )

    def __init__(
        self,
        name: str,
        protocol_version: Optional[str] = None,
        language_version: Optional[str] = None,
        speed: Optional[float] = None,
        file_revision: Optional[str] = None,
        channel_name: Optional[str] = None,
    ):
        super().__init__()
        self.name = name
        self.protocol_version = protocol_version
        self.language_version = language_version
        self.file_revision = file_revision
        self.speed = speed
        self.channel_name = channel_name

    def __str__(self) -> str:
        return (
            f'LinNetwork(name = "{self.name}", protocol_version = "{self.protocol_version}", language_version = "{self.language_version}",'
            f'file_revision = "{self.file_revision or ""}", speed = {self.speed}, channel_name = "{self.channel_name or ""}")'
        )

    __repr__ = __str__


class LinNode(Node):
    """ """

    lin_node_id = Column(Integer, ForeignKey("node.rid"), primary_key=True)

    signals = relationship("LinSignal", back_populates="publisher")

    __mapper_args__ = {"polymorphic_identity": "LinNode"}


class LinMasterNode(LinNode):
    """ """

    lin_master_node_id = Column(
        Integer, ForeignKey("linnode.lin_node_id"), primary_key=True
    )

    timebase = StdFloat(default=0.0)
    jitter = StdFloat(default=0.0)

    # Only J2602.
    bit_length = StdInteger(nullable=True, default=None)
    tolerant = StdFloat(nullable=True, default=None)

    __mapper_args__ = {"polymorphic_identity": "LinMasterNode"}

    def __init__(self, name, timebase, jitter, bit_length=None, tolerant=None):
        self.name = name
        self.timebase = timebase
        self.jitter = jitter
        self.bit_length = bit_length
        self.tolerant = tolerant

    def __repr__(self):
        return 'LinMasterNode(name = "{}", timebase = {}, jitter = {}, bit_length = {}, tolerant = {})'.format(
            self.name, self.timebase, self.jitter, self.bit_length, self.tolerant
        )

    __str__ = __repr__


class LinSlaveNode(LinNode):
    """ """

    lin_slave_node_id = Column(
        Integer, ForeignKey("linnode.lin_node_id"), primary_key=True
    )

    protocol_version = Column(Unicode(255))
    configured_NAD = StdInteger(nullable=True, default=None)
    initial_NAD = StdInteger(nullable=True, default=None)

    supplier_id = StdInteger(default=0)
    function_id = StdInteger(default=0)
    variant = StdInteger(nullable=True, default=None)

    p2_min = StdFloat(nullable=True, default=None)
    st_min = StdFloat(nullable=True, default=None)
    n_as_timeout = StdFloat(nullable=True, default=None)
    n_cr_timeout = StdFloat(nullable=True, default=None)
    response_tolerance = StdFloat(nullable=True, default=None)

    @hybrid_property
    def product_id(self):
        return LinProductIdType(self.supplier_id, self.function_id, self.variant)

    @product_id.setter
    def product_id(self, product_id):
        """ """
        if product_id in ("", ()):
            self.supplier_id = self.function_id = self.variant = 0
        elif isinstance(product_id, LinProductIdType):
            self.supplier_id = product_id.supplier_id
            self.function_id = product_id.function_id
            self.variant = product_id.variant
        elif isinstance(product_id, str):
            parts = product_id.split(".")
            self.supplier_id, self.function_id, self.variant = [int(c) for c in parts]
        elif isinstance(product_id, tuple):
            if len(product_id) != 3:
                raise ValueError(
                    "product_id as tuple must contain three elements: (supplier_id, function_id, variant)"
                )
            else:
                supplier_id, function_id, variant = product_id
                self.supplier_id = int(supplier_id)
                self.function_id = int(function_id)
                self.variant = int(variant)
        else:
            raise TypeError("product_id must be LinProductIdType, string, or 3-tuple.")

    __mapper_args__ = {"polymorphic_identity": "LinSlaveNode"}

    def __init__(
        self,
        name,
        protocol_version=None,
        configured_NAD=None,
        initial_NAD=None,
        product_id=(),
        p2_min=None,
        st_min=None,
        n_as_timeout=None,
        n_cr_timeout=None,
        response_tolerance=None,
    ):
        self.name = name
        self.protocol_version = protocol_version
        self.configured_NAD = configured_NAD
        self.initial_NAD = initial_NAD
        self.product_id = product_id
        self.p2_min = p2_min
        self.st_min = st_min
        self.n_as_timeout = n_as_timeout
        self.n_cr_timeout = n_cr_timeout
        self.response_tolerance = response_tolerance

    def __repr__(self):
        return 'LinSlaveNode(name = "{}", protocol_version = "{}", configured_NAD = {}, initial_NAD = {}, product_id = {}, p2_min = {}, st_min = {}, n_as_timeout = {}, n_cr_timeout = {}, response_tolerance = {})'.format(
            self.name,
            self.protocol_version,
            self.configured_NAD,
            self.initial_NAD,
            self.product_id,
            self.p2_min,
            self.st_min,
            self.n_as_timeout,
            self.n_cr_timeout,
            self.response_tolerance,
        )

    __str__ = __repr__


class LinSignal(Signal):
    """LIN (Local Interconnect Network) signal model"""

    lin_signal_id: Mapped[int] = mapped_column(
        Integer, ForeignKey("signal.rid"), primary_key=True
    )
    _init_value: Mapped[str] = mapped_column(Unicode(255), default="0")

    publisher_id: Mapped[int] = mapped_column(
        Integer,
        ForeignKey("linnode.lin_node_id", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
    )
    publisher: Mapped["LinNode"] = relationship("LinNode", back_populates="signals")

    signal_subscribers: Mapped[List["LinSignalSubscriber"]] = relationship(
        "LinSignalSubscriber", cascade="all, delete-orphan", back_populates="signal"
    )
    subscribers = association_proxy("signal_subscribers", "subscriber")

    __mapper_args__ = {"polymorphic_identity": "LinSignal"}

    def __init__(
        self,
        name: str,
        signal_size: int,
        init_value: Union[int, List[int]],
        publisher: "LinNode",
    ):
        super().__init__()
        self.name = name
        self.bitsize = signal_size
        self.byteorder = ByteOrderType.MOTOROLA
        self.valuetype = ValueType.INT
        self.init_value = init_value
        self.publisher = publisher

    @hybrid_property
    def init_value(self) -> Union[int, List[int]]:
        """Get the initial value of the signal"""
        value = self._init_value
        match = INIT_VALUE.match(value)
        if not match:
            raise ValueError(f"Malformed initial value '{value}'.")

        gd = match.groupdict()
        match gd:
            case {"array": array} if array:
                return [int(v) for v in value.split(",")]
            case {"scalar": scalar} if scalar:
                return int(value)
            case _:
                raise ValueError(f"Malformed initial value '{value}'.")

    @init_value.setter
    def init_value(self, value: Union[int, List[int], tuple]) -> None:
        """Set the initial value of the signal"""
        match value:
            case int():
                self._init_value = str(value)
            case list() | tuple():
                self._init_value = ",".join([str(x) for x in value])
            case _:
                raise TypeError(
                    f"Expected int, list, or tuple, got {type(value).__name__}"
                )

    def __str__(self) -> str:
        return f'LinSignal(name = "{self.name}", signal_size = {self.bitsize}, init_value = {self.init_value})'

    __repr__ = __str__


class LinSignalSubscriber(Base, MixInBase):
    """ """

    lin_node_id = Column(
        Integer,
        ForeignKey("linnode.lin_node_id", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        primary_key=True,
    )
    lin_signal_id = Column(
        Integer,
        ForeignKey("linsignal.lin_signal_id", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        primary_key=True,
    )
    subscriber = relationship(LinNode, lazy="joined")
    signal = relationship(
        "LinSignal", back_populates="signal_subscribers", lazy="joined"
    )

    def __init__(self, subscriber):
        self.subscriber = subscriber


class LinUnconditionalFrame(Message):
    """ """

    lin_unconditional_frame_id = Column(
        Integer, ForeignKey("message.rid"), primary_key=True
    )

    frame_signals = relationship(
        "LinUnconditionalFrameSignal", cascade="all, delete-orphan", backref="frame"
    )
    signals = association_proxy("frame_signals", "signal")

    publisher_id = Column(
        Integer,
        ForeignKey("linnode.lin_node_id", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
    )
    publisher = relationship("LinNode", backref="unconditional_frames")

    __mapper_args__ = {"polymorphic_identity": "LinUnconditionalFrame"}

    def __init__(self, name, frame_id, size, publisher):
        self.name = name
        self.size = size
        self.frame_id = frame_id
        self.publisher = publisher

    def __repr__(self):
        return 'LinUnconditionalFrame(name = "{}", frame_id = {}, size = {}, publisher = {})'.format(
            self.name, self.frame_id, self.size, self.publisher
        )

    __str__ = __repr__

    @hybrid_property
    def size(self):
        return self.dlc

    @size.setter
    def size(self, value):
        if not (1 <= value <= 8):
            raise ValueError("'size' must be in range [1 .. 8]")
        self.dlc = value

    @hybrid_property
    def frame_id(self):
        return self.message_id

    @frame_id.setter
    def frame_id(self, value):
        if not (0 <= value <= 59):
            raise ValueError("'frame_id' must be in range [0 .. 59]")
        self.message_id = value

    def add_signal(self, signal, signal_offset):
        return self.frame_signals.append(
            LinUnconditionalFrameSignal(signal, signal_offset)
        )


class LinUnconditionalFrameSignal(Base, MixInBase):
    """ """

    lin_unconditional_frame_id = Column(
        Integer,
        ForeignKey("linunconditionalframe.lin_unconditional_frame_id"),
        primary_key=True,
    )

    lin_signal_id = Column(
        Integer,
        ForeignKey("linsignal.lin_signal_id", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        primary_key=True,
    )
    signal_offset = StdInteger(default=0)
    unconditional_frame = relationship(
        LinUnconditionalFrame, lazy="joined", overlaps="frame,frame_signals"
    )
    signal = relationship(LinSignal, lazy="joined")

    def __init__(self, signal, signal_offset=0):
        self.signal = signal
        self.signal_offset = signal_offset

    def __repr__(self):
        return "LinUnconditionalFrameSignal(unconditional_frame = {}, signal = {}, signal_offset = {})".format(
            self.unconditional_frame, self.signal, self.signal_offset
        )

    __str__ = __repr__


class LinSignalEncodingType(Base, RidMixIn):
    """ """

    name = Column(Unicode(255), nullable=False, unique=True)

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return 'LinSignalEncodingType(name = "{}")'.format(self.name)

    __str__ = __repr__


class LinSignalRepresentation(Base, MixInBase):
    """ """

    lin_signal_encoding_type_id = Column(
        Integer,
        ForeignKey("linsignalencodingtype.rid"),
        nullable=False,
        primary_key=True,
    )
    signal_id = Column(
        Integer,
        ForeignKey("signal.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    signal_encoding_type = relationship(
        "LinSignalEncodingType", backref="representations", uselist=False
    )
    signal = relationship("Signal", uselist=False)

    def __init__(self, signal_encoding_type, signal):
        self.signal_encoding_type = signal_encoding_type
        self.signal = signal

    def __repr__(self):
        return "LinSignalRepresentation(signal_encoding_type = {}, signal = {})".format(
            self.signal_encoding_type, self.signal
        )

    __str__ = __repr__


class LinSignalEncodingEntry(Base, MixInBase):
    entry_id = Column(Integer, primary_key=True)

    lin_signal_encoding_type_id = Column(
        Integer,
        ForeignKey("linsignalencodingtype.rid"),
        nullable=False,  # , primary_key = True
    )
    type = Column(String(75))
    signal_encoding_type = relationship(
        "LinSignalEncodingType", backref="entries", uselist=False
    )

    __mapper_args__ = {
        "polymorphic_identity": "LinSignalEncodingEntry",
        "polymorphic_on": type,
    }


class LinSignalEncodingEntry_Value(LinSignalEncodingEntry):
    entry_id = Column(
        Integer, ForeignKey("linsignalencodingentry.entry_id"), primary_key=True
    )
    entry_type = Column(Integer)

    __mapper_args__ = {"polymorphic_identity": "LinSignalEncodingEntry_Value"}

    def __init__(self, entry_type):
        self.entry_type = entry_type

    def __repr__(self):
        return "LinSignalEncodingEntry_Value(entry_type = {})".format(self.entry_type)

    __str__ = __repr__


class LinSignalEncodingEntry_Logical(LinSignalEncodingEntry):
    entry_id = Column(
        Integer, ForeignKey("linsignalencodingentry.entry_id"), primary_key=True
    )
    signal_value = Column(Float, nullable=False)
    text_info = Column(Unicode(1024), nullable=True)

    __mapper_args__ = {"polymorphic_identity": "LinSignalEncodingEntry_Logical"}

    def __init__(self, signal_value, text_info):
        self.signal_value = signal_value
        self.text_info = text_info

    def __repr__(self):
        return 'LinSignalEncodingEntry_Logical(signal_value = {}, text_info = "{}")'.format(
            self.signal_value, self.text_info
        )

    __str__ = __repr__


class LinSignalEncodingEntry_Physical(LinSignalEncodingEntry):
    entry_id = Column(
        Integer, ForeignKey("linsignalencodingentry.entry_id"), primary_key=True
    )
    min_value = Column(Float, nullable=True)
    max_value = Column(Float, nullable=True)
    scale = Column(Float, nullable=True)
    offset = Column(Float, nullable=True)
    text_info = Column(Unicode(1024), nullable=True)

    __mapper_args__ = {"polymorphic_identity": "LinSignalEncodingEntry_Physical"}

    def __init__(self, min_value, max_value, scale, offset, text_info):
        min_value, max_value, scale, offset, text_info

    def __repr__(self):
        return 'LinSignalEncodingEntry_Physical(min_value = {}, max_value = {}, scale = {}, offset = {}, text_info = "{}")'.format(
            self.min_value, self.max_value, self.scale, self.offset, self.text_info
        )

    __str__ = __repr__


class LinScheduleTable(Base, RidMixIn):
    """ """

    name = Column(Unicode(255), nullable=False, unique=True)
    commands = relationship(
        "LinScheduleTable_Command",
        order_by="LinScheduleTable_Command.position",
        collection_class=ordering_list("position"),
    )


class LinScheduleTable_Command(Base, MixInBase):
    """ """

    rid = Column(Integer, primary_key=True, autoincrement=True)

    lin_schedule_table_id = Column(
        Integer, ForeignKey("linscheduletable.rid"), nullable=False
    )
    frame_time = Column(Float, nullable=False)
    type = Column(String(255))
    position = StdInteger()

    __mapper_args__ = {
        "polymorphic_identity": "LinScheduleTable_Command",
        "polymorphic_on": type,
    }

    def __init__(self, frame_time):
        self.frame_time = frame_time

    def __repr__(self):
        return "LinScheduleTable_Command(frame_time = {})".format(self.frame_time)

    __str__ = __repr__


class LinScheduleTable_Command_Frame(LinScheduleTable_Command):
    """ """

    command_id = Column(
        Integer, ForeignKey("linscheduletable_command.rid"), primary_key=True
    )
    frame_id = Column(
        Integer,
        ForeignKey("linunconditionalframe.lin_unconditional_frame_id"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    frame = relationship("LinUnconditionalFrame", backref="entries", uselist=False)

    __mapper_args__ = {"polymorphic_identity": "LinScheduleTable_Command_Frame"}

    def __init__(self, frame_time, frame):
        self.frame_time = frame_time
        self.frame = frame

    def __repr__(self):
        return "LinScheduleTable_Command_Frame(frame_time = {}, frame = {})".format(
            self.frame_time, self.frame
        )

    __str__ = __repr__


class LinScheduleTable_Command_MasterReq(LinScheduleTable_Command):
    """ """

    command_id = Column(
        Integer, ForeignKey("linscheduletable_command.rid"), primary_key=True
    )

    __mapper_args__ = {"polymorphic_identity": "LinScheduleTable_Command_MasterReq"}

    def __init__(self, frame_time):
        self.frame_time = frame_time

    def __repr__(self):
        return "LinScheduleTable_Command_MasterReq(frame_time = {})".format(
            self.frame_time
        )

    __str__ = __repr__


class LinScheduleTable_Command_SlaveResp(LinScheduleTable_Command):
    """ """

    command_id = Column(
        Integer, ForeignKey("linscheduletable_command.rid"), primary_key=True
    )

    __mapper_args__ = {"polymorphic_identity": "LinScheduleTable_Command_SlaveResp"}

    def __init__(self, frame_time):
        self.frame_time = frame_time

    def __repr__(self):
        return "LinScheduleTable_Command_SlaveResp(frame_time = {})".format(
            self.frame_time
        )

    __str__ = __repr__


class LinScheduleTable_Command_AssignNad(LinScheduleTable_Command):
    """ """

    command_id = Column(
        Integer, ForeignKey("linscheduletable_command.rid"), primary_key=True
    )
    node_id = Column(
        Integer,
        ForeignKey("node.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    node = relationship("Node", backref="assign_nads", uselist=False)

    __mapper_args__ = {"polymorphic_identity": "LinScheduleTable_Command_AssignNad"}

    def __init__(self, frame_time, node):
        self.frame_time = frame_time
        self.node = node

    def __repr__(self):
        return "LinScheduleTable_Command_AssignNad(frame_time = {}, node = {})".format(
            self.frame_time, self.node
        )

    __str__ = __repr__


class LinScheduleTable_Command_ConditionalChangeNad(LinScheduleTable_Command):
    """ """

    command_id = Column(
        Integer, ForeignKey("linscheduletable_command.rid"), primary_key=True
    )
    nad = Column(Integer)
    id = Column(Integer)
    byte = Column(Integer)
    mask = Column(Integer)
    inv = Column(Integer)
    new_nad = Column(Integer)

    __mapper_args__ = {
        "polymorphic_identity": "LinScheduleTable_Command_ConditionalChangeNad"
    }

    def __init__(self, frame_time, nad, id, byte, mask, inv, new_nad):
        self.frame_time = frame_time
        self.nad = nad
        self.id = id
        self.byte = byte
        self.mask = mask
        self.inv = inv
        self.new_nad = new_nad

    def __repr__(self):
        return "LinScheduleTable_Command_ConditionalChangeNad(frame_time = {}, nad = {}, id = {}, byte = {}, mask = {}, inv = {}, new_nad = {})".format(
            self.frame_time,
            self.nad,
            self.id,
            self.byte,
            self.mask,
            self.inv,
            self.new_nad,
        )

    __str__ = __repr__


class LinScheduleTable_Command_DataDump(LinScheduleTable_Command):
    """ """

    command_id = Column(
        Integer, ForeignKey("linscheduletable_command.rid"), primary_key=True
    )
    node_id = Column(
        Integer,
        ForeignKey("node.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    node = relationship("Node", backref="data_dumps", uselist=False)
    d1 = Column(Integer)
    d2 = Column(Integer)
    d3 = Column(Integer)
    d4 = Column(Integer)
    d5 = Column(Integer)

    __mapper_args__ = {"polymorphic_identity": "LinScheduleTable_Command_DataDump"}

    def __init__(self, frame_time, node, d1, d2, d3, d4, d5):
        self.frame_time = frame_time
        self.node = node
        self.d1 = d1
        self.d2 = d2
        self.d3 = d3
        self.d4 = d4
        self.d5 = d5

    def __repr__(self):
        return "LinScheduleTable_Command_DataDump(frame_time = {}, node = {}, d1 = {}, d2 = {}, d3 = {}, d4 = {}, d5 = {})".format(
            self.frame_time, self.node, self.d1, self.d2, self.d3, self.d4, self.d5
        )

    __str__ = __repr__


class LinScheduleTable_Command_SaveConfiguration(LinScheduleTable_Command):
    """ """

    command_id = Column(
        Integer, ForeignKey("linscheduletable_command.rid"), primary_key=True
    )
    node_id = Column(
        Integer,
        ForeignKey("node.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    node = relationship("Node", backref="save_configurations", uselist=False)

    __mapper_args__ = {
        "polymorphic_identity": "LinScheduleTable_Command_SaveConfiguration"
    }

    def __init__(self, frame_time, node):
        self.frame_time = frame_time
        self.node = node

    def __repr__(self):
        return "LinScheduleTable_Command_SaveConfiguration(frame_time = {}, node = {})".format(
            self.frame_time, self.node
        )

    __str__ = __repr__


class LinScheduleTable_Command_AssignFrameIdRange(LinScheduleTable_Command):
    """ """

    command_id = Column(
        Integer, ForeignKey("linscheduletable_command.rid"), primary_key=True
    )
    node_id = Column(
        Integer,
        ForeignKey("node.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    node = relationship("Node", backref="assign_frame_id_ranges", uselist=False)
    frame_index = Column(Integer)
    frame_pid1 = Column(Integer)
    frame_pid2 = Column(Integer)
    frame_pid3 = Column(Integer)
    frame_pid4 = Column(Integer)

    __mapper_args__ = {
        "polymorphic_identity": "LinScheduleTable_Command_AssignFrameIdRange"
    }

    def __init__(
        self,
        frame_time,
        node,
        frame_index,
        frame_pid1=None,
        frame_pid2=None,
        frame_pid3=None,
        frame_pid4=None,
    ):
        self.frame_time = frame_time
        self.node = node
        self.frame_index = frame_index
        self.frame_pid1 = frame_pid1
        self.frame_pid2 = frame_pid2
        self.frame_pid3 = frame_pid3
        self.frame_pid4 = frame_pid4

    def __repr__(self):
        return "LinScheduleTable_Command_AssignFrameIdRange(frame_time = {}, node = {}, frame_index = {}, frame_pid1 = {}, frame_pid2 = {}, frame_pid3 = {}, frame_pid4 = {})".format(
            self.frame_time,
            self.node,
            self.frame_index,
            self.frame_pid1,
            self.frame_pid2,
            self.frame_pid3,
            self.frame_pid4,
        )

    __str__ = __repr__


class LinScheduleTable_Command_FreeFormat(LinScheduleTable_Command):
    """ """

    command_id = Column(
        Integer, ForeignKey("linscheduletable_command.rid"), primary_key=True
    )
    d1 = Column(Integer)
    d2 = Column(Integer)
    d3 = Column(Integer)
    d4 = Column(Integer)
    d5 = Column(Integer)
    d6 = Column(Integer)
    d7 = Column(Integer)
    d8 = Column(Integer)

    __mapper_args__ = {"polymorphic_identity": "LinScheduleTable_Command_FreeFormat"}

    def __init__(self, frame_time, d1, d2, d3, d4, d5, d6, d7, d8):
        self.frame_time = frame_time
        self.d1 = d1
        self.d2 = d2
        self.d3 = d3
        self.d4 = d4
        self.d5 = d5
        self.d6 = d6
        self.d7 = d7
        self.d8 = d8

    def __repr__(self):
        return "LinScheduleTable_Command_FreeFormat(frame_time = {}, d1 = {}, d2 = {}, d3 = {}, d4 = {}, d5 = {}, d6 = {}, d7 = {}, d8 = {})".format(
            self.frame_time,
            self.d1,
            self.d2,
            self.d3,
            self.d4,
            self.d5,
            self.d6,
            self.d7,
            self.d8,
        )

    __str__ = __repr__


class LinScheduleTable_Command_AssignFrameId(LinScheduleTable_Command):
    """ """

    command_id = Column(
        Integer, ForeignKey("linscheduletable_command.rid"), primary_key=True
    )

    node_id = Column(
        Integer,
        ForeignKey("node.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        # primary_key=True,
    )
    node = relationship("Node", backref="assign_frame_ids", uselist=False)
    frame_id = Column(
        Integer,
        ForeignKey("linunconditionalframe.lin_unconditional_frame_id"),
        nullable=False,
        default=0,
        # primary_key=True,
    )
    frame = relationship(
        "LinUnconditionalFrame", backref="assign_frame_ids", uselist=False
    )

    __mapper_args__ = {"polymorphic_identity": "LinScheduleTable_Command_AssignFrameId"}

    def __init__(self, node: Node, frame: LinUnconditionalFrame, frame_time: int):
        self.node = node
        self.frame = frame
        self.frame_time = frame_time

    def __repr__(self):
        return "LinScheduleTable_Command_AssignFrameId(frame_time = {}, node = {}, frame = {})".format(
            self.frame_time, self.node, self.frame
        )

    __str__ = __repr__


class LinSporadicFrame_Association(Base, MixInBase):
    """ """

    sporadic_frame_id = Column(
        Integer, ForeignKey("linsporadicframe.frame_id"), primary_key=True
    )
    frame_id = Column(
        Integer,
        ForeignKey("linunconditionalframe.lin_unconditional_frame_id"),
        primary_key=True,
    )
    unconditional_frame = relationship("LinUnconditionalFrame")
    position = StdInteger()

    def __init__(self, unconditional_frame):
        self.unconditional_frame = unconditional_frame


class LinSporadicFrame(LinUnconditionalFrame):
    """ """

    frame_id = Column(
        Integer,
        ForeignKey("linunconditionalframe.lin_unconditional_frame_id"),
        primary_key=True,
    )
    associated_frames = relationship(
        "LinUnconditionalFrame",
        secondary="linsporadicframe_association",
        uselist=True,
        overlaps="unconditional_frame",
    )

    """
    message_signals = relationship(
        "Message_Signal", cascade="all, delete-orphan", backref="message"
    )
    signals = association_proxy("message_signals", "signal")
    """

    __mapper_args__ = {"polymorphic_identity": "LinSporadicFrame"}

    def __init__(self, name):
        self.name = name


class LinEventTriggeredFrame_Association(Base, MixInBase):
    """ """

    event_triggered_id = Column(
        Integer, ForeignKey("lineventtriggeredframe.frame_id"), primary_key=True
    )
    frame_id = Column(
        Integer,
        ForeignKey("linunconditionalframe.lin_unconditional_frame_id"),
        primary_key=True,
    )
    event_triggered_frame = relationship(
        "LinEventTriggeredFrame", foreign_keys=[event_triggered_id]
    )
    unconditional_frame = relationship("LinUnconditionalFrame")

    # position = StdInteger()

    def __init__(self, unconditional_frame):
        self.unconditional_frame = unconditional_frame


class LinEventTriggeredFrame(LinUnconditionalFrame):
    """ """

    frame_id = Column(
        Integer,
        ForeignKey("linunconditionalframe.lin_unconditional_frame_id"),
        primary_key=True,
    )
    collision_resolving_schedule_table_id = Column(
        Integer,
        ForeignKey("linscheduletable.rid"),
        nullable=True,
    )

    collision_resolving_schedule_table = relationship(
        "LinScheduleTable", backref="event_triggered_frames", uselist=False
    )

    associated_frames = relationship(
        "LinUnconditionalFrame",
        secondary="lineventtriggeredframe_association",
        order_by="LinEventTriggeredFrame_Association.frame_id",
        # collection_class = ordering_list('position'),
        uselist=True,
        overlaps="unconditional_frame,event_triggered_frame",
    )

    # associated_frames_assoc = relationship("LinEventTriggeredFrame_Association",
    #    order_by = "LinEventTriggeredFrame_Association.position", collection_class = ordering_list('position'),
    #    #backref = "event_triggered_frame"
    # )
    # associated_frames = association_proxy("associated_frames_assoc", "event_triggered_frame")

    __mapper_args__ = {"polymorphic_identity": "LinEventTriggeredFrame"}

    def __init__(
        self, name, frame_id, master_node, collision_resolving_schedule_table=None
    ):
        self.name = name
        self.frame_id = frame_id
        self.publisher = master_node
        self.collision_resolving_schedule_table = collision_resolving_schedule_table


class LinConfigurableFrame(Base, MixInBase):
    """ """

    node_id = Column(Integer, ForeignKey("node.rid"), nullable=False, primary_key=True)
    frame_id = Column(
        Integer,
        ForeignKey("linunconditionalframe.lin_unconditional_frame_id"),
        primary_key=True,
    )
    identifier = StdInteger(nullable=True)

    node = relationship("Node", backref="configurable_frames", uselist=False)
    frame = relationship(
        "LinUnconditionalFrame", backref="configurable_frames", uselist=False
    )


class LinFaultStateSignal(Base, MixInBase):
    """ """

    node_id = Column(Integer, ForeignKey("node.rid"), nullable=False, primary_key=True)
    signal_id = Column(
        Integer,
        ForeignKey("signal.rid", onupdate="CASCADE", ondelete="RESTRICT"),
        nullable=False,
        default=0,
        primary_key=True,
    )
    node = relationship("Node", backref="fault_state_signals", uselist=False)
    signal = relationship("Signal", uselist=False)

    def __init__(self, node, signal):
        self.node = node
        self.signal = signal

    def __str__(self):
        return "LinFaultStateSignal(node = {}, signal = {})".format(
            self.node, self.signal
        )

    __repr__ = __str__


class LinResponseErrorSignal(Base, MixInBase):
    """ """

    node_id = Column(
        Integer, ForeignKey("linnode.lin_node_id"), nullable=False, primary_key=True
    )

    signal_id = Column(
        Integer, ForeignKey("linsignal.lin_signal_id"), nullable=True, primary_key=True
    )

    node = relationship("LinNode", backref="response_error", uselist=False)
    signal = relationship("LinSignal", uselist=False)

    def __init__(self, node, signal):
        self.node = node
        self.signal = signal

    def __str__(self):
        return "LinNodeResponseError(node = {}, signal = {})".format(
            self.node, self.signal
        )

    __repr__ = __str__
